from typing import Union

from pika import BasicProperties
from pika.exceptions import ChannelClosedByBroker

from petisco.base.domain.message.command import Command
from petisco.base.domain.message.command_bus import CommandBus
from petisco.extra.rabbitmq.application.message.configurer.rabbitmq_message_configurer import (
    RabbitMqMessageConfigurer,
)
from petisco.extra.rabbitmq.application.message.consumer.rabbitmq_consumer_connector import (
    RabbitMqConsumerConnector,
)
from petisco.extra.rabbitmq.application.message.formatter.rabbitmq_message_queue_name_formatter import (
    RabbitMqMessageQueueNameFormatter,
)
from petisco.extra.rabbitmq.shared.rabbitmq_connector import RabbitMqConnector


class RabbitMqCommandBus(CommandBus):
    def __init__(
        self,
        organization: str,
        service: str,
        connector: Union[
            RabbitMqConnector, RabbitMqConsumerConnector
        ] = RabbitMqConnector(),
    ):
        self.connector = connector
        self.exchange_name = f"{organization}.{service}"
        self.rabbitmq_key = f"publisher-{self.exchange_name}"
        self.configurer = RabbitMqMessageConfigurer(organization, service, connector)
        self.properties = BasicProperties(delivery_mode=2)  # PERSISTENT_TEXT_PLAIN

    def dispatch(self, command: Command):
        self._check_is_command(command)
        meta = self.get_configured_meta()
        command = command.update_meta(meta)
        try:
            channel = self.connector.get_channel(self.rabbitmq_key)
            routing_key = RabbitMqMessageQueueNameFormatter.format(
                command, exchange_name=self.exchange_name
            )
            channel.confirm_delivery()
            channel.basic_publish(
                exchange=self.exchange_name,
                routing_key=routing_key,
                body=command.json(),
                properties=self.properties,
            )
            if channel.is_open and not isinstance(
                self.connector, RabbitMqConsumerConnector
            ):
                channel.close()
        except ChannelClosedByBroker:
            # If domain event queue is not configured, it will be configured and then try to publish again.
            self.configurer.configure()
            self.publish(command)

    def close(self):
        self.connector.close(self.rabbitmq_key)



from typing import Dict, Type

from petisco.base.application.controller.http_error import HttpError
from petisco.base.domain.errors.domain_error import DomainError

ErrorMap = Dict[Type[DomainError], HttpError]



import json
import re
from datetime import datetime
from typing import Dict

from petisco.base.domain.model.uuid import Uuid
from petisco.base.domain.model.value_object import ValueObject

TIME_FORMAT = "%Y-%m-%d %H:%M:%S.%f"


def get_version(config) -> int:
    version = getattr(config, "version", 1) if config else 1
    return version


def get_message_name(namespace) -> str:
    return (
        re.sub(r"(?<!^)(?=[A-Z])", "_", namespace.get("__qualname__", "message"))
        .lower()
        .replace("_", ".")
    )


class MetaMessage(type):
    def __new__(mcs, name, bases, namespace):
        config = namespace.get("Config")

        namespace["version"] = get_version(config)
        namespace["name"] = get_message_name(namespace)
        namespace["attributes"] = {}
        namespace["meta"] = {}

        return super().__new__(mcs, name, bases, namespace)


class Message(metaclass=MetaMessage):
    def __init__(self, **kwargs):
        self.message_id: Uuid
        self.name: str
        self.version: int
        self.occurred_on: datetime
        self.attributes: Dict = {}
        self.meta: Dict = {}
        self.type: str = "message"
        self._set_data(**kwargs)

    def _set_data(self, **kwargs):
        if kwargs:
            self.message_id = (
                Uuid.from_value(kwargs.get("id")) if kwargs.get("id") else Uuid.v4()
            )
            self.name = kwargs.get("type")
            self.version = kwargs.get("version")
            self.occurred_on = (
                datetime.strptime(kwargs.get("occurred_on"), TIME_FORMAT)
                if kwargs.get("occurred_on")
                else datetime.now()
            )
            self.attributes = kwargs.get("attributes")
            self.meta = kwargs.get("meta")
            self.type = kwargs.get("type_message", "message")
        else:
            self.message_id = Uuid.v4()
            self.occurred_on = datetime.utcnow()

    def _set_attributes(self, **kwargs):
        if self.attributes is None:
            self.attributes = {}
        for k in kwargs:
            self.attributes[k] = kwargs[k]

    def add_meta(self, meta: Dict):
        self.meta = meta

    def update_meta(self, meta: Dict):
        if not meta:
            return self

        if not isinstance(meta, Dict):
            raise TypeError("Message.update_meta() expect a dict")
        if self.meta:
            self.meta = {**self.meta, **meta}
        else:
            self.meta = meta
        return self

    @staticmethod
    def from_dict(message_data: Dict):
        data = message_data.get("data")
        return Message(**data)

    @staticmethod
    def from_json(message_json: str):
        event_dict = json.loads(message_json)
        return Message.from_dict(event_dict)

    def _get_serialized_attributes(self) -> Dict:
        attributes = {}
        for key, attribute in self.attributes.items():
            serialized_value = attribute
            if isinstance(attribute, ValueObject):
                serialized_value = attribute.value
            if isinstance(attribute, datetime):
                serialized_value = attribute.strftime(TIME_FORMAT)
            attributes[key] = serialized_value
        return attributes

    def dict(self) -> Dict:
        data = {
            "data": {
                "id": self.message_id.value,
                "type": self.name,
                "type_message": self.type,
                "version": self.version,
                "occurred_on": self.occurred_on.strftime(TIME_FORMAT),
                "attributes": self._get_serialized_attributes(),
                "meta": self.meta,
            }
        }
        return data

    def to_str(self, class_name="Message", type: str = "message"):
        return f"{class_name} [{self.message_id.value} ({type}), {self.name} (v{self.version}), {self.occurred_on}, attributes={self.attributes}]"

    def json(self):
        return json.dumps(self.dict())

    def __eq__(self, other):
        return self.dict() == other.dict()

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return self.to_str()



from typing import Dict

from petisco.base.application.notifier.notifier_exception_message import (
    NotifierExceptionMessage,
)
from petisco.extra.slack.application.notifier.create_text_meta import create_text_meta
from petisco.extra.slack.application.notifier.slack_notifier_message_converter import (
    SlackNotifierMessageConverter,
)


class ExceptionBlocksSlackNotifierMessageConverter(SlackNotifierMessageConverter):
    def __init__(
        self,
        slack_accessory: Dict = None,
        repository_url: str = None,
        repository_name_button: str = "Code",
    ):
        self.slack_accessory = slack_accessory
        self.repository_url = repository_url
        self.repository_name_button = repository_name_button

    def _create_header_block(self, title: str):
        return {
            "type": "header",
            "text": {
                "type": "plain_text",
                "text": title,
                "emoji": True,
            },
        }

    def _create_message_block(
        self, notifier_exception_message: NotifierExceptionMessage
    ):
        text_meta = create_text_meta(notifier_exception_message.meta)
        message_block = None
        if text_meta:
            message_block = {
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": text_meta,
                },
            }
            if self.slack_accessory:
                message_block["accessory"] = self.slack_accessory
        return message_block

    def _create_executor_block(
        self, notifier_exception_message: NotifierExceptionMessage
    ):
        executor_block = {
            "type": "section",
        }

        if self.repository_url:
            url_error = f"{self.repository_url}/{notifier_exception_message.filename}"
            if notifier_exception_message.lineno:
                url_error += f"#L{notifier_exception_message.lineno}"
            executor_block["accessory"] = {
                "type": "button",
                "text": {
                    "type": "plain_text",
                    "text": self.repository_name_button,
                    "emoji": True,
                },
                "value": "click_me_repository_code",
                "url": url_error,
                "action_id": "button-repository-code",
                "style": "danger",
            }

        text_executor = f"*Executor*: {notifier_exception_message.executor}"
        if notifier_exception_message.filename:
            text_executor += f"\n {notifier_exception_message.filename}"
        if notifier_exception_message.lineno:
            text_executor += f" | Line {notifier_exception_message.lineno}"

        if notifier_exception_message.input_parameters:
            input_parameters_text = ""
            for k, v in notifier_exception_message.input_parameters.items():
                input_parameters_text += f"\n>{k}: {v}"
            text_executor += input_parameters_text[: 3000 - len(text_executor)]

        executor_block["text"] = {
            "type": "mrkdwn",
            "text": text_executor,
        }
        return executor_block

    def _create_error_block(self, notifier_exception_message: NotifierExceptionMessage):
        error_block = None
        if notifier_exception_message.exception:
            text_error = (
                f"*Class*: {notifier_exception_message.exception.__class__}\n"
                f">{str(notifier_exception_message.exception)[:500]}"
            )
            traceback_base_str = "\n```{}```"
            traceback_max_length = 3000 - len(text_error) - len(traceback_base_str)
            text_error += traceback_base_str.format(
                notifier_exception_message.traceback[:traceback_max_length]
            )

            error_block = {
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": text_error,
                },
            }
        return error_block

    def convert(self, notifier_exception_message: NotifierExceptionMessage):

        blocks = []
        header_block = self._create_header_block(notifier_exception_message.title)
        message_block = self._create_message_block(notifier_exception_message)
        executor_block = self._create_executor_block(notifier_exception_message)
        error_block = self._create_error_block(notifier_exception_message)

        blocks.append(header_block)
        if message_block:
            blocks.append(message_block)
        blocks.append(executor_block)

        if error_block:
            blocks.append(error_block)

        divider_block = {"type": "divider"}
        blocks.append(divider_block)

        return blocks



from abc import abstractmethod
from typing import Any

from meiga import Error, NotImplementedMethodError, Result

from petisco.base.application.controller.controller import Controller
from petisco.extra.fastapi.controller.fastapi_result_mapper import FastAPIResultMapper


class FastAPIController(Controller):
    @staticmethod
    def get_default_mapper():
        return FastAPIResultMapper.default()

    @staticmethod
    def get_config_mapper(config):
        return FastAPIResultMapper.from_config(config)

    @abstractmethod
    def execute(self, *args, **kwargs) -> Result[Any, Error]:
        return NotImplementedMethodError



from meiga import Result

from petisco import __version__
from petisco.base.application.application_info import ApplicationInfo
from petisco.base.application.dependency_injection.container import Container
from petisco.base.application.middleware.middleware import Middleware
from petisco.base.application.notifier.notifier_exception_message import (
    NotifierExceptionMessage,
)
from petisco.base.application.notifier.notifier_message import NotifierMessage
from petisco.base.domain.errors.critical_error import CriticalError
from petisco.base.domain.errors.unknown_error import UnknownError


class NotifierMiddleware(Middleware):
    def __init__(self, wrapped_class_name, wrapped_class_input_arguments):
        super().__init__(wrapped_class_name, wrapped_class_input_arguments)
        self.notifier = Container.get("notifier")

    def before(self):
        pass

    def after(self, result: Result):
        if result.is_failure:
            error = result.value

            meta = {
                "petisco": __version__,
                "application_name": ApplicationInfo().name,
                "application_version": ApplicationInfo().version,
            }
            if issubclass(error.__class__, UnknownError):
                notifier_exception_message = (
                    NotifierExceptionMessage.from_unknown_error(
                        error, title="Uncontrolled Exception"
                    )
                )
                notifier_exception_message.update_meta(meta)
                self.notifier.publish_exception(notifier_exception_message)
            if issubclass(error.__class__, CriticalError):
                notifier_message = NotifierMessage(
                    title=error.get_specify_detail(),
                    message=error.__repr__(),
                    meta=meta,
                )
                self.notifier.publish(notifier_message)



from typing import Dict

from meiga import Error


class DomainError(Error):
    def __init__(self, uuid_value: str = None, additional_info: Dict[str, str] = None):
        self.uuid_value = uuid_value
        self.additional_info = additional_info
        self._additional_detail = ""
        self._specific_detail = "DomainError"
        self._set_detail()

    def _set_detail(self):
        if self.uuid_value:
            self._additional_detail += f" ({self.uuid_value})"

        if self.additional_info:
            self._additional_detail += f" [{self.additional_info}]"

    @classmethod
    def get_specify_detail(cls):
        return cls.__name__

    def detail(self) -> str:
        return f"{self.get_specify_detail()}{self._additional_detail}"

    def __str__(self):
        return f"{self.get_specify_detail()}{self._additional_detail}"

    def __repr__(self):
        return f"{self.get_specify_detail()}{self._additional_detail}"



import argparse
import logging
from time import sleep
from typing import List, Type

from meiga import BoolResult, isFailure, isSuccess

from petisco import (
    AllMessageSubscriber,
    DomainEvent,
    Message,
    MessageSubscriber,
    __version__,
)
from petisco.extra.rabbitmq import RabbitMqConnector, RabbitMqMessageConsumer
from petisco.legacy import LoggingBasedLogger

ORGANIZATION = "alice"
RETRY_TTL = 5000  # default
MAX_RETRIES = 5  # default


def has_args(args):
    is_active = False
    for arg in vars(args):
        is_active = is_active or getattr(args, arg)
    return is_active


def get_logger():
    def logging_config():
        logging.getLogger("pika").setLevel(logging.WARNING)

    logger = LoggingBasedLogger("example", config=logging_config)
    return logger


class MyDomainEvent(DomainEvent):
    ...


class UnackMessage(MessageSubscriber):
    def subscribed_to(self) -> List[Type[Message]]:
        return [MyDomainEvent]

    def handle(self, message: Message) -> BoolResult:
        print(f"> It will unack the message: {message.dict()}\n")
        return isFailure


def get_args():
    parser = argparse.ArgumentParser(
        prog="petisco-rabbitmq 🍪",
        description="petisco-rabbitmq helps us on rabbitmq iteration",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument("-rq", "--requeue", action="store_true", help="requeue")
    parser.add_argument(
        "-o",
        "--organization",
        action="store",
        dest="organization",
        default=ORGANIZATION,
        help="Name of the organization",
    )
    parser.add_argument(
        "-s",
        "--service",
        action="store",
        dest="service",
        default=None,
        help="Name of the service",
    )
    parser.add_argument(
        "-cq",
        "--consuming-queue",
        action="store",
        dest="consuming_queue",
        default=None,
        help="Queue to consume",
    )
    parser.add_argument(
        "-rrk",
        "--retry-routing-key",
        action="store",
        dest="retry_routing_key",
        default=None,
        help="Routing key to republish the message to specific retry queue",
    )
    parser.add_argument(
        "-ren",
        "--retry-exchange-name",
        action="store",
        dest="retry_exchange_name",
        default=None,
        help="Exchange name to republish the message to specific exchange",
    )
    parser.add_argument(
        "-mr",
        "--max-retries",
        action="store",
        dest="max_retries",
        default=MAX_RETRIES,
        help="Max Retries",
    )
    parser.add_argument(
        "-rttl",
        "--retry-ttl",
        action="store",
        dest="retry_ttl",
        default=RETRY_TTL,
        help="Retry TTL",
    )
    parser.add_argument(
        "-wtr",
        "--wait-to-requeue",
        action="store",
        dest="wait_to_requeue",
        default=None,
        help="Wait to Requeue (seconds)",
    )

    args = parser.parse_args()

    if not has_args(args):
        parser.print_help()
        return None

    return args


def main():
    args = get_args()

    if not args:
        return

    if args.requeue:

        if not args.consuming_queue:
            print(" ➜ Not consuming queue given")
            return

        if not args.retry_routing_key:
            print(" ➜ Not retry routing_key given")
            return

        if not args.service:
            print(" ➜  Not service provided")
            return

        print(f"petisco-rabbitmq ({__version__}) 🍪 ")
        print(f"🍪 Requeuing events from: {args.consuming_queue}")
        print(f"🍪 With the following routing_key: {args.retry_routing_key}")
        if args.retry_exchange_name:
            print(f"🍪 With retry exchange name: {args.retry_exchange_name}")
        print(f"🍪 ORGANIZATION {args.organization}")
        print(f"🍪 SERVICE {args.service}")
        print(f"🍪 MAX_RETRIES {args.max_retries}")
        print(f"🍪 WAIT_TO_REQUEUE {args.wait_to_requeue} seconds")

        connector = RabbitMqConnector()
        consumer = RabbitMqMessageConsumer(
            args.organization,
            args.service,
            args.max_retries,
            connector,
            logger=get_logger(),
        )

        class RequeueOnMessage(AllMessageSubscriber):
            def handle(self, message: Message) -> BoolResult:
                if args.wait_to_requeue:
                    sleep(args.wait_sec)

                self.domain_event_bus.retry_publish(
                    message, args.retry_routing_key, args.retry_exchange_name
                )

                return isSuccess

        consumer.add_subscriber_on_queue(
            queue_name=args.consuming_queue,
            subscriber=RequeueOnMessage,
            message_type_expected="domain_event",
        )

        consumer.start()



from petisco.extra.rabbitmq.dependencies import get_default_message_dependencies
from petisco.extra.rabbitmq.is_pika_available import is_pika_available

__all__ = ["get_default_message_dependencies"]


if is_pika_available():

    from petisco.extra.rabbitmq.application.chaos.rabbitmq_message_chaos import (
        RabbitMqMessageChaos,
    )
    from petisco.extra.rabbitmq.application.message.bus.rabbitmq_command_bus import (
        RabbitMqCommandBus,
    )
    from petisco.extra.rabbitmq.application.message.bus.rabbitmq_domain_event_bus import (
        RabbitMqDomainEventBus,
    )
    from petisco.extra.rabbitmq.application.message.configurer.rabbitmq_message_configurer import (
        RabbitMqMessageConfigurer,
    )
    from petisco.extra.rabbitmq.application.message.consumer.rabbitmq_message_consumer import (
        RabbitMqMessageConsumer,
    )
    from petisco.extra.rabbitmq.dependencies import get_rabbitmq_message_dependencies
    from petisco.extra.rabbitmq.rabbitmq_message_application_configurer import (
        RabbitMqMessageApplicationConfigurer,
    )
    from petisco.extra.rabbitmq.shared.queue_config import QueueConfig
    from petisco.extra.rabbitmq.shared.rabbitmq_connector import RabbitMqConnector
    from petisco.extra.rabbitmq.shared.rabbitmq_declarer import RabbitMqDeclarer
    from petisco.extra.rabbitmq.shared.rabbitmq_is_running_locally import (
        rabbitmq_is_running_locally,
    )
    from petisco.extra.rabbitmq.shared.specific_queue_config import SpecificQueueConfig

    __all__ += [
        "RabbitMqConnector",
        "RabbitMqDeclarer",
        "RabbitMqDomainEventBus",
        "RabbitMqMessageConfigurer",
        "QueueConfig",
        "SpecificQueueConfig",
        "RabbitMqMessageConsumer",
        "RabbitMqCommandBus",
        "get_rabbitmq_message_dependencies",
        "RabbitMqMessageChaos",
        "rabbitmq_is_running_locally",
        "RabbitMqMessageApplicationConfigurer",
    ]



from typing import List

from petisco.base.domain.message.message_configurer import MessageConfigurer
from petisco.base.domain.message.message_subscriber import MessageSubscriber


class NotImplementedMessageConfigurer(MessageConfigurer):
    def configure_subscribers(
        self,
        subscribers: List[MessageSubscriber],
        clear_subscriber_before: bool = False,
        clear_store_before: bool = False,
    ):
        pass

    def clear(self):
        pass



from uuid import uuid4

import validators
from pydantic import validator

from petisco.base.domain.errors.defaults.invalid_uuid import InvalidUuid
from petisco.base.domain.model.value_object import ValueObject


class Uuid(ValueObject):
    @validator("value")
    def validate_value(cls, value):
        if value is None or not validators.uuid(value):
            raise InvalidUuid(uuid_value=value)
        return value

    @classmethod
    def v4(cls):
        return cls(value=str(uuid4()))
